#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
import utool as ut


if __name__ == '__main__':
    """
    SeeAlso:
        utool.util_autogen.makeinit
    """
    ut.execute_doctest(ut.makeinit, testnum=0)
