#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
autogenerates sphinx documentation
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import utool as ut


if __name__ == '__main__':
    """
    autogen_sphinx_apidoc
    OR
    python -m utool.util_setup --exec-autogen_sphinx_apidoc
    """
    ut.autogen_sphinx_apidoc()
