# -*- coding: utf-8 -*-
"""
Aliases harder to remember or very common standard module functions
"""
from __future__ import absolute_import, division, print_function

from itertools import product, combinations
from collections import defaultdict, OrderedDict, namedtuple
from functools import partial

ddict = defaultdict
odict = OrderedDict
iprod = product
icomb = combinations
namedtuple = namedtuple
partial = partial
