# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import six
global_cache_fname = 'global_cache.shelf'
global_cache_dname = 'global_cache' + ('' if six.PY2 else '_py3')
default_appname = 'utool'
